from __future__ import annotations

import pytest

from skoufas_dbf_reader.field_extractors import *


def test_has_author():
    assert not has_author(None)
    assert not has_author("")
    assert has_author("BLA")
    assert not has_author("ΧΣ")
    assert not has_author("X,S")
    assert not has_author("X.S.")
    assert not has_author("X.S")
    assert not has_author("X . Σ.")
    assert not has_author("X.Σ.")
    assert not has_author("X.Σ")
    assert not has_author("Χ.Σ")
    assert not has_author("Χ.Σ.")
    assert not has_author("Χ. Σ .")
    assert not has_author("Χ,Σ")
    assert not has_author("Χ. Σ.")
    assert not has_author("Χ.Χ")
    assert not has_author("Χ.Χ.")


def test_language_codes():
    assert language_codes()
    assert type(language_codes()) == dict
    assert len(language_codes()) > 0


def test_has_language():
    assert not has_language(None)
    assert not has_language("")
    assert not has_language("GAL")
    assert not has_language("ΨΑΡΟΜΗΛΙΓΚΟΣ , ΛΑΖΟΥ , ΚΑΡΤΑΛΗΣ ")
    assert has_language("ΩUENEAU RAYMOND                    GAL")
    assert not has_language("ΚΕΛΕΣΙΔΗΣ, ΤΕΛΗΣ                       Ι")
    assert has_language("BITSIOS,DIMITRIS                  AGL")
    assert has_language("FINLAY GEORG                        GER")


def test_authors_from_a01():
    assert authors_from_a01(None) == []
    assert authors_from_a01("") == []
    assert authors_from_a01("X,S") == []
    assert authors_from_a01("GAL") == ["GAL"]
    assert authors_from_a01("ΨΑΡΟΜΗΛΙΓΚΟΣ , ΛΑΖΟΥ , ΚΑΡΤΑΛΗΣ ") == ["ΨΑΡΟΜΗΛΙΓΚΟΣ,", "ΛΑΖΟΥ,", "ΚΑΡΤΑΛΗΣ,"]
    assert authors_from_a01("ΚΕΛΕΣΙΔΗΣ, ΤΕΛΗΣ                       Ι") == ["ΚΕΛΕΣΙΔΗΣ,ΤΕΛΗΣ"]
    assert authors_from_a01("ΩUENEAU RAYMOND                    GAL") == ["QUENEAU,RAYMOND"]
    assert authors_from_a01("BITSIOS,DIMITRIS                  AGL") == ["ΒΙΤΣΙΟΣ,ΔΗΜΗΤΡΗΣ"]
    assert authors_from_a01("FINLAY GEORG                        GER") == ["FINLAY,GEORGE"]
    assert authors_from_a01("ΑΝΑΓΝΩΣΤΑΚΗΣ,ΗΛΙΑΣ    .") == ["ΑΝΑΓΝΩΣΤΑΚΗΣ,ΗΛΙΑΣ"]


def test_language_from_a01():
    assert not language_from_a01(None)
    assert not language_from_a01("")
    assert not language_from_a01("GAL")
    assert not language_from_a01("ΨΑΡΟΜΗΛΙΓΚΟΣ , ΛΑΖΟΥ , ΚΑΡΤΑΛΗΣ ")
    assert not language_from_a01("ΚΕΛΕΣΙΔΗΣ, ΤΕΛΗΣ                       Ι")
    assert language_from_a01("ΩUENEAU RAYMOND                    GAL") == "FR"
    assert language_from_a01("BITSIOS,DIMITRIS                  AGL") == "EN"
    assert language_from_a01("FINLAY GEORG                        GER") == "DE"


def test_title_from_a02():
    assert title_from_a02(None) is None
    assert title_from_a02("") is None
    assert title_from_a02(" ") is None
    assert title_from_a02("GAL ") == "GAL"
    assert title_from_a02(" GAL") == "GAL"


def test_subtitle_from_a03():
    assert subtitle_from_a03(None) is None
    assert subtitle_from_a03("") is None
    assert subtitle_from_a03(" ") is None
    assert subtitle_from_a03("GAL ") == "GAL"
    assert subtitle_from_a03(" GAL") == "GAL"


def test_dewey_from_a04_a05():
    assert dewey_from_a04_a05(None, None) is None
    assert dewey_from_a04_a05("", " ") is None
    assert dewey_from_a04_a05(" ", "") is None
    assert dewey_from_a04_a05("001A", "") == "001 A"
    assert dewey_from_a04_a05("001.009ΚΟΝ ", "") == "001.009 ΚΟΝ"
    assert dewey_from_a04_a05(" 001.009ΚΟΝ", "") == "001.009 ΚΟΝ"
    assert dewey_from_a04_a05("O.50ΠΕΡ", "") == "050 ΠΕΡ"
    assert dewey_from_a04_a05("624.183Χ.Σ", "") == "624.183 ΧΣ"
    assert dewey_from_a04_a05("320ΤΣΟ", "") == "320 ΤΣΟ"

    assert dewey_from_a04_a05("HOEMANN", "") is None


def test_entry_numbers_from_a04_a05_a06_a07_a08_a18_a19():
    assert entry_numbers_from_a04_a05_a06_a07_a08_a18_a19(None, None, None, None, None, None, None) == []
    assert entry_numbers_from_a04_a05_a06_a07_a08_a18_a19("", "", "", "", "", "", "") == []
    assert entry_numbers_from_a04_a05_a06_a07_a08_a18_a19("", " ", None, "", "", "", "") == []
    assert entry_numbers_from_a04_a05_a06_a07_a08_a18_a19(None, None, "", "", "", "", "") == []
    assert entry_numbers_from_a04_a05_a06_a07_a08_a18_a19("8123", "2710-2709", "", "", "", "", "") == [
        "8123",
        "2710",
        "2709",
    ]
    assert entry_numbers_from_a04_a05_a06_a07_a08_a18_a19("098.122 DEW", "2710-2709", None, None, None, None, None) == [
        "2710",
        "2709",
    ]
    assert entry_numbers_from_a04_a05_a06_a07_a08_a18_a19(None, "2710-2709", "foobar", "baz", "yum", "foo", "bar") == [
        "2710",
        "2709",
    ]
    assert entry_numbers_from_a04_a05_a06_a07_a08_a18_a19(None, "2710-2710", "foobar", "baz", "yum", "foo", "bar") == [
        "2710",
    ]
    assert entry_numbers_from_a04_a05_a06_a07_a08_a18_a19(
        "Some Dewey", "2710-2709", "-", "baz", "yum", "lol", "lal"
    ) == ["2710", "2709"]
    assert entry_numbers_from_a04_a05_a06_a07_a08_a18_a19(
        "", "2710-2709", "-10450", "baz", "yum", "ooo", "0980980"
    ) == [
        "2710",
        "2709",
        "10450",
    ]
    assert entry_numbers_from_a04_a05_a06_a07_a08_a18_a19(
        "", "2710-2709", "1747-1746-1745-", "baz", "yum", "34", "1414241"
    ) == [
        "2710",
        "2709",
        "1747",
        "1746",
        "1745",
    ]
    assert entry_numbers_from_a04_a05_a06_a07_a08_a18_a19("", "10-1", "4", "baz", "yum", "34", "1414241") == [
        "10",
        "14",
    ]
    assert entry_numbers_from_a04_a05_a06_a07_a08_a18_a19("", "10-1", "5", "baz", "yum", "34", "1414241") == [
        "10",
        "15",
    ]
    assert entry_numbers_from_a04_a05_a06_a07_a08_a18_a19(
        "", "10-1", "448 ΚΩΣΤΑΣ ΦΙΛΙΝΗΣ", "baz", "yum", "34", "1414241"
    ) == [
        "10",
        "1448",
    ]
    assert entry_numbers_from_a04_a05_a06_a07_a08_a18_a19(
        "", "4225-4226-4228-4229-", "4227-4290-4536-4535-", "4537", "ΕΥΡΩΠΑΙΚ.ΚΕΝΤΡ.ΤΕΧΝΗ", "34", "1414241"
    ) == ["4225", "4226", "4228", "4229", "4227", "4290", "4536", "4535", "4537"]
    assert entry_numbers_from_a04_a05_a06_a07_a08_a18_a19(
        "", "5280-5285-5286-5283-", "5284-5278-5277-5279-", "5281-", "5282-6548-6547", "asasa", "1414241"
    ) == "5280-5285-5286-5283-5284-5278-5277-5279-5281-5282-6548-6547".split("-")

    assert entry_numbers_from_a04_a05_a06_a07_a08_a18_a19(
        "",
        "1938-1937-1936-1935-",
        "1933-1870-1869-1866-",
        "1932",
        "ΣΑΚΕΛΛΑΡΙΟΥ",
        "1931-1867-",
        "1868-1934",
    ) == "1938-1937-1936-1935-1933-1870-1869-1866-1932-1931-1867-1868-1934".split("-")


def test_translator_from_a06():
    assert translator_from_a06(None) is None
    assert translator_from_a06("") is None
    assert translator_from_a06(" ") is None
    assert translator_from_a06("hello") == "hello"
    # Part of field06_corrections
    assert translator_from_a06("-") is None
    assert translator_from_a06("1203") is None
    assert translator_from_a06("3ΕΚΔ") is None
    assert translator_from_a06("2847 ΠΑΠΑΡΡΟΔΟΥ,ΝΙΚ") == "ΠΑΠΑΡΡΟΔΟΥ,ΝΙΚΟΛΑΟΣ"


def test_edition_from_a07():
    assert edition_from_a07(None) is None
    assert edition_from_a07("") is None
    assert edition_from_a07(" ") is None
    assert edition_from_a07("2") == "2"
    assert edition_from_a07("6ΕΚΔ") == "6"
    # invalid
    assert edition_from_a07("ΠΡΟΣΚΗ") is None
    # year
    assert edition_from_a07("1932") is None
    # series
    assert edition_from_a07("4537") is None


def test_editor_from_a08_a09():
    assert editor_from_a08_a09(None, None) is None
    assert editor_from_a08_a09("", "") is None
    assert editor_from_a08_a09(" ", " ") is None
    assert editor_from_a08_a09("hello", " ") == ("hello", None)
    assert editor_from_a08_a09("", "city") == (None, "city")
    assert editor_from_a08_a09("5282-6548-6547", "city") == (None, "city")
    assert editor_from_a08_a09('"Η ΔΑΜΑΣΚΟΣ"', "city") == ("Η ΔΑΜΑΣΚΟΣ", "city")
    assert editor_from_a08_a09("X.E", "city") == (None, "city")
    assert editor_from_a08_a09("hello", "1867") == ("hello", None)
    assert editor_from_a08_a09("hello", "AMSTERDAN") == ("hello", "AMSTERDAM")
    assert editor_from_a08_a09("hello", "ΑΘΗΝΑ 1984") == ("hello", "ΑΘΗΝΑ")
    assert editor_from_a08_a09("hello", "X.T") == ("hello", None)
    assert editor_from_a08_a09("Α.Α. ΛΙΒΑΝΗ", None) == ("ΛΙΒΑΝΗ", "ΑΘΗΝΑ")


def test_edition_year_from_a09_a10():
    assert edition_year_from_a09_a10(None, None) is None
    assert edition_year_from_a09_a10("", None) is None
    assert edition_year_from_a09_a10(" ", "") is None
    assert edition_year_from_a09_a10("", "Χ.Τ") is None
    assert edition_year_from_a09_a10("", "ΑΡΤΑ") is None
    assert edition_year_from_a09_a10("", "Α989") == 1989
    assert edition_year_from_a09_a10("", "MCMX") == 1910
    assert edition_year_from_a09_a10("", "2010") == 2010
    assert edition_year_from_a09_a10("", "197Ο") == 1970
    assert edition_year_from_a09_a10("1867", "197Ο") == 1867

    with pytest.raises(Exception) as e_info:
        edition_year_from_a09_a10("", "foobar")
    assert e_info.exconly() == "Exception: invalid date [foobar]"


def test_pages_from_a11():
    assert pages_from_a11(None) is None
    assert pages_from_a11("") is None
    assert pages_from_a11(" ") is None
    with pytest.raises(Exception) as e_info:
        pages_from_a11("foobar")
    assert e_info.exconly() == "Exception: invalid pages [foobar]"
    assert pages_from_a11("127Σ") == 127
    assert pages_from_a11("447Σ Ε") == 447
    assert pages_from_a11("256ΣΕΓ") == 256
    assert pages_from_a11("29ΙΣ") == 291


def test_topics_from_a12_to_a15_a20_a22_to_a24():
    assert topics_from_a12_to_a15_a20_a22_to_a24(None) == []
    assert topics_from_a12_to_a15_a20_a22_to_a24([]) == []
    assert topics_from_a12_to_a15_a20_a22_to_a24(["", ""]) == []
    assert topics_from_a12_to_a15_a20_a22_to_a24(["\\"]) == []
    assert topics_from_a12_to_a15_a20_a22_to_a24(["'"]) == []
    assert topics_from_a12_to_a15_a20_a22_to_a24(["A", "B"]) == ["A", "B"]
    assert topics_from_a12_to_a15_a20_a22_to_a24(["A-B", "B"]) == ["A", "B"]
    assert topics_from_a12_to_a15_a20_a22_to_a24(["19 ΑΙΩΝΑ", "B"]) == ["19 ΑΙΩΝΑΣ", "B"]
    assert topics_from_a12_to_a15_a20_a22_to_a24(["foobar-(12-13)"]) == ["12-13", "foobar"]
    assert topics_from_a12_to_a15_a20_a22_to_a24(["foobar(12-13)"]) == ["12-13", "foobar"]
    assert topics_from_a12_to_a15_a20_a22_to_a24(["foo bar(12-13)"]) == ["12-13", "foo bar"]
    assert topics_from_a12_to_a15_a20_a22_to_a24(["foo-bar-(12-13)"]) == ["12-13", "foo", "bar"]
    assert topics_from_a12_to_a15_a20_a22_to_a24(["foo bar(1213)"]) == ["1213", "foo bar"]
    assert topics_from_a12_to_a15_a20_a22_to_a24(["foo bar(1213qw)"]) == ["1213qw", "foo bar"]
    assert topics_from_a12_to_a15_a20_a22_to_a24(["foobar-(19 ΑΙΩΝΑ)"]) == ["19 ΑΙΩΝΑΣ", "foobar"]


def test_curator_from_a16():
    assert curator_from_a16(None) is None
    assert curator_from_a16("") is None
    assert curator_from_a16(" ") is None
    assert curator_from_a16("woo") == "woo"
    assert curator_from_a16("ΘΕΣΣΑΛΙΑ") is None
    assert curator_from_a16("#") is None
    assert curator_from_a16("ΒΡΑΣΑΣ,ΝΙΚΟΣ") == "ΒΡΑΣΣΑΣ,ΝΙΚΟΛΑΟΣ"


def test_has_cd_from_a02_a03_a12_a13_a14_a17_a18_a22_a30():
    assert not has_cd_from_a02_a03_a12_a13_a14_a17_a18_a22_a30(None)
    assert not has_cd_from_a02_a03_a12_a13_a14_a17_a18_a22_a30([None, "", "", "gooh", "wordwithcdinside hello"])
    assert has_cd_from_a02_a03_a12_a13_a14_a17_a18_a22_a30(["", "ΠΕΡΙΕΧΕΙ CD"])
    assert has_cd_from_a02_a03_a12_a13_a14_a17_a18_a22_a30(["", "ΤΟΜΟΙ Α-Β-Γ  2 ΑΝΤΙΤΥΠΑ ΠΕΡΙΕΧΕΙ CD"])
    assert has_cd_from_a02_a03_a12_a13_a14_a17_a18_a22_a30(["", "ΤΟΜΟΙ Α-Β-Γ  2 ΑΝΤΙΤΥΠΑ ΠΕΡΙΕΧΕΙ CD"])


def test_has_dvd_from_a30():
    assert not has_dvd_from_a30(None)
    assert not has_dvd_from_a30([None, "", "", "gooh", "wordwithDVDinside hello"])
    assert has_dvd_from_a30(["", "ΠΕΡΙΕΧΕΟ DVD"])


def test_copies_from_a17_a18_a30():
    assert copies_from_a17_a18_a30(None, None, None) is None
    assert copies_from_a17_a18_a30("weewv", "ecwecwc", "woo") is None
    assert copies_from_a17_a18_a30("ΑΝΤΙΤΥΠΑ 2", None, None) == 2
    assert copies_from_a17_a18_a30(None, None, "5 ΑΝΤΙΤΥΠΑ") == 5
    assert copies_from_a17_a18_a30(None, "2ΑΝΤΙΤΥΠΑ", "qefqef") == 2
    assert copies_from_a17_a18_a30(None, None, "ΑΝΑΤΥΠΟ") is None


def test_donation_from_a17_a30():
    assert donation_from_a17_a30(None, None) is None
    assert donation_from_a17_a30("weewv", "ecwecwc") is None
    assert donation_from_a17_a30("ΒΙΒΙΟΘΗΚΗ ΓΑΡΟΥΦΑΛΙΑ", None) == "ΒΙΒΛΙΟΘΗΚΗ ΓΑΡΟΥΦΑΛΙΑ"
    assert donation_from_a17_a30(None, "ΔΩΡΑ ΟΛΓΑ ΜΑΝΟΥ") == "ΜΑΝΟΥ,ΟΛΓΑ"
    assert donation_from_a17_a30(None, "ΑΝΑΤΥΠΟ") is None


def test_offprint_from_a17_a30():
    assert not offprint_from_a17_a21_a30(None, None, None)
    assert not offprint_from_a17_a21_a30("", "", "")
    assert not offprint_from_a17_a21_a30("weewv", "foo", "ecwecwc")
    assert offprint_from_a17_a21_a30("weewv", "", "ΑΝΑΤΥΠΟ")
    assert offprint_from_a17_a21_a30("ΑΝΑΤΥΠΟ", "foo", "")
    assert offprint_from_a17_a21_a30(" wfe", "ΑΝΑΤΥΠΟ", "foo")


def test_volume_from_a17_a18_a20_a30():
    assert volume_from_a17_a18_a20_a30(None, None, None, None) is None
    assert volume_from_a17_a18_a20_a30("", "", "", "") is None
    assert volume_from_a17_a18_a20_a30("wffw", "wef", "wef", "wef") is None
    assert volume_from_a17_a18_a20_a30("", "ΤΕΥΧΗ 10", "", "") == "ΤΕΥΧΗ 10"


def test_material_from_a18_a30():
    assert material_from_a18_a30(None, None) is None
    assert material_from_a18_a30("", "") is None
    assert material_from_a18_a30("aaa", "aa") is None
    assert material_from_a18_a30("ΧΑΡΤ", "") == "ΧΑΡΤΕΣ"
    assert material_from_a18_a30("fwfwe", "ΠΕΡΙΕΧΕΙ 12 ΧΑΡΤΕΣ") == "12 ΧΑΡΤΕΣ"


def test_notes_from_a17_a18_a21_a30():
    assert notes_from_a17_a18_a21_a30(None, None, None, None) is None
    assert notes_from_a17_a18_a21_a30("", "", "", "") is None
    assert notes_from_a17_a18_a21_a30("  ", None, "", "") is None
    assert (
        notes_from_a17_a18_a21_a30(
            "foo",
            "ignored",
            "bar",
            "baz",
        )
        == "foo\nbar\nbaz"
    )
    assert (
        notes_from_a17_a18_a21_a30(
            "2607-2 ΑΝΤΙΤΥΠΑ",
            "ignored",
            "bar",
            "baz",
        )
        == "2607\nbar\nbaz"
    )
    assert (
        notes_from_a17_a18_a21_a30("foo", "ignored", "", "2 ΑΝΤΙΤΥΠΑ ΤΟ ΕΝΑ ΞΕΝΟΓΛΩΣΣΟ")
        == "foo\n2 ΑΝΤΙΤΥΠΑ ΤΟ ΕΝΑ ΞΕΝΟΓΛΩΣΣΟ"
    )
    assert (
        notes_from_a17_a18_a21_a30("foo", "0518-2867", "", "2 ΑΝΤΙΤΥΠΑ ΤΟ ΕΝΑ ΞΕΝΟΓΛΩΣΣΟ")
        == "foo\n0518-2867\n2 ΑΝΤΙΤΥΠΑ ΤΟ ΕΝΑ ΞΕΝΟΓΛΩΣΣΟ"
    )


def test_isbn_from_a17_a18_a19__a22_a30():
    assert isbn_from_a17_a18_a19_a22_a30(None, None, None, None, None) is None
    assert isbn_from_a17_a18_a19_a22_a30("", "", "", "", "") is None
    assert isbn_from_a17_a18_a19_a22_a30(" ", " ", " ", " ", " ") is None
    assert isbn_from_a17_a18_a19_a22_a30("", None, " ", None, "") is None
    assert isbn_from_a17_a18_a19_a22_a30(" 16 ΤΟΜΟΙ", "9789605039", "332", "foo", None) == "9789605039332"
    assert isbn_from_a17_a18_a19_a22_a30("bla", None, None, "blu", "bla") is None
    assert isbn_from_a17_a18_a19_a22_a30("12 TOMOI", "8440909306", None, None, "ΔΩΡΕΑ Ι.ΠΝΕΥΜΑΤΙΚΟΥ") == "8440909306"
    assert isbn_from_a17_a18_a19_a22_a30("2 ΑΝΤΙΤΥΠΑ", "960-05-034", "3-5", None, None) == "960-05-0343-5"
    assert isbn_from_a17_a18_a19_a22_a30("2 ΑΝΤΙΤΥΠΑ", "960-70-25-", "64-4", None, None) == "960-70-25-64-4"
    assert isbn_from_a17_a18_a19_a22_a30("2 ΑΝΤΙΤΥΠΑ", "978-455-44", "3-0", None, None) == "978-455-443-0"
    assert isbn_from_a17_a18_a19_a22_a30("2 ΑΝΤΙΤΥΠΑ", "978-8778-1", "670-40", None, None) == "978-8778-1670-40"
    assert isbn_from_a17_a18_a19_a22_a30("2 ΑΝΤΙΤΥΠΑ", "978-960-20", "8-813-5", None, None) == "978-960-208-813-5"
    assert isbn_from_a17_a18_a19_a22_a30("2 ΑΝΤΙΤΥΠΑ", "9789603031", "574", None, None) == "9789603031574"
    assert isbn_from_a17_a18_a19_a22_a30("2 ΤΟΜΟΙ", "960-14-005", "6-7", None, None) == "960-14-0056-7"
    assert isbn_from_a17_a18_a19_a22_a30("2 ΤΟΜΟΙ", "960-14-012", "7-Χ", None, None) == "960-14-0127-Χ"
    assert isbn_from_a17_a18_a19_a22_a30("2 ΤΟΜΟΙ", "960239269Χ", None, None, None) == "960239269Χ"
    assert isbn_from_a17_a18_a19_a22_a30("4 ΤΕΥΧΗ", "17927099", None, None, "ΔΩΡΕΑ ΟΛΓΑ ΜΑΝΟΥ") == "17927099"
    assert isbn_from_a17_a18_a19_a22_a30("960-2565-440-7", None, None, None, None) == "960-2565-440-7"
    assert isbn_from_a17_a18_a19_a22_a30("9786180123951", None, None, None, None) == "9786180123951"
    assert isbn_from_a17_a18_a19_a22_a30("9789601419985", None, None, None, None) == "9789601419985"
    assert isbn_from_a17_a18_a19_a22_a30("9789605273941", None, None, None, None) == "9789605273941"
    assert isbn_from_a17_a18_a19_a22_a30("9789606051456", None, None, None, None) == "9789606051456"
    assert (
        isbn_from_a17_a18_a19_a22_a30(
            "ΔΩΡΕΑ ΚΩΝ/ΝΟΥ ΤΣΙΛΙΓΙΑΝΝΗ",
            "960-6601-6",
            "8-4",
            "",
            "                                                      960-6601-68",
        )
        == "960-6601-68-4"
    )
    assert isbn_from_a17_a18_a19_a22_a30("ΔΩΡΕΑ ΓΙΑΝΝΗ ΜΠΑΝΙΑ", None, None, None, "960-88458-3-1") == "960-88458-3-1"
    assert (
        isbn_from_a17_a18_a19_a22_a30("ΔΩΡΕΑ ΔΗΜΗΤΡΙΟΥ ΠΟΡΤΟΑΒΑΡΑ", None, None, None, "978-960-503-483-2")
        == "978-960-503-483-2"
    )

    # assert isbn_from_a17_a18_a19_a22_a30("foo", None, "7027-05-1", "5-6", None) is None
    assert isbn_from_a17_a18_a19_a22_a30("foo", None, "7027-05-1", "5-6", None) == "7027-05-15-6"



from __future__ import annotations

import os
from collections import defaultdict

import pytest
import yaml

from skoufas_dbf_reader.field_extractors import *
from skoufas_dbf_reader.utilities import all_entries, check_ean, check_isbn, check_issn, is_valid_dewey_strict, romanize


@pytest.fixture
def reports_directory() -> str:
    r = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir, "reports"))
    os.makedirs(r, exist_ok=True)
    return r


def minimal_entry(input_entry: dict[int, str], fields: list[int]) -> dict[int, str]:
    return {i: input_entry[i] for i in fields if input_entry[i]}


def test_report_single_fields(reports_directory: str):
    field_values: list[set[str]] = [set() for _ in range(31)]
    for entry in all_entries():
        for i in range(1, 31):
            if i in entry and entry[i] and entry[i].strip():
                field_values[i].add(entry[i].strip())
    for i in range(1, 31):
        with open(os.path.join(reports_directory, f"fields_{i:02}.yml"), "w", encoding="utf-8") as outfile:
            yaml.dump(list(sorted(field_values[i])), outfile, default_flow_style=False, allow_unicode=True)


def test_report_single_extracted_fields(reports_directory: str):
    field_values: defaultdict[str, list[str | list[str]]] = defaultdict(list)
    for entry in all_entries():
        authors = authors_from_a01(entry[1])
        for author in authors:
            field_values["author"].append(author)
            if plain_author_re.fullmatch(author) or author in author_corrections().values():
                field_values["plain_author"].append(author)
            else:
                field_values["weird_author"].append(author)

        language = language_from_a01(entry[1])
        if language:
            field_values["language"].append(language)

        title = title_from_a02(entry[2])
        if title:
            field_values["title"].append(title)

        subtitle = subtitle_from_a03(entry[3])
        if subtitle:
            field_values["subtitle"].append(subtitle)

        dewey = dewey_from_a04_a05(entry[4], entry[5])
        if dewey:
            field_values["dewey"].append(dewey)

        entry_numbers = entry_numbers_from_a04_a05_a06_a07_a08_a18_a19(
            entry[4], entry[5], entry[6], entry[7], entry[8], entry[18], entry[19]
        )

        field_values["entry_number_lists"].append(entry_numbers)
        for entry_number in entry_numbers:
            field_values["entry_numbers"].append(entry_number)

        translator = translator_from_a06(entry[6])
        if translator:
            translators = translator.split("!!")
            for single_translator in translators:
                field_values["translator"].append(single_translator)
                translator_surname_name = single_translator.split(",", maxsplit=1)
                if len(translator_surname_name) == 2:
                    field_values["translator_family_name"].append(translator_surname_name[0])
                    if translator_surname_name[1].endswith("."):
                        field_values["translator_name_abbreviations"].append(translator_surname_name[1])
                    else:
                        field_values["translator_names"].append(translator_surname_name[1])

        edition = edition_from_a07(entry[7])
        if edition:
            field_values["edition"].append(edition)

        editor = editor_from_a08_a09(entry[8], entry[9])
        if editor:
            if not editor[0] or not editor[1]:
                field_values["editor"].append(f"{editor[0]} // {editor[1]} ({entry[0]})")
            else:
                field_values["editor"].append(f"{editor[0]} // {editor[1]}")

        edition_year = edition_year_from_a09_a10(entry[9], entry[10])
        if edition_year:
            field_values["edition_year"].append(str(edition_year))

        pages = pages_from_a11(entry[11])
        if pages:
            field_values["pages"].append(str(pages))

        topic_list = topics_from_a12_to_a15_a20_a22_to_a24(
            [
                entry[12],
                entry[13],
                entry[14],
                entry[15],
                entry[20],
                entry[22],
                entry[23],
                entry[24],
            ]
        )
        field_values["topic_lists"].append(topic_list)
        for topic in topic_list:
            field_values["topics"].append(topic)

        curator = curator_from_a16(entry[16])
        if curator:
            field_values["curator"].append(curator)

        copies = copies_from_a17_a18_a30(entry[17], entry[18], entry[30])
        if copies:
            field_values["copies"].append(str(copies))

        donation = donation_from_a17_a30(entry[17], entry[30])
        if donation:
            field_values["donation"].append(donation)

        volume = volume_from_a17_a18_a20_a30(entry[17], entry[18], entry[20], entry[30])
        if volume:
            field_values["volume"].append(volume)

        material = material_from_a18_a30(entry[18], entry[30])
        if material:
            field_values["material"].append(material)

        notes = notes_from_a17_a18_a21_a30(entry[17], entry[18], entry[21], entry[30])
        if notes:
            field_values["notes"].append(notes)

        isbn = isbn_from_a17_a18_a19_a22_a30(entry[17], entry[18], entry[19], entry[22], entry[30])
        if isbn:
            if not check_isbn(isbn):
                field_values["isbn"].append(isbn)
            if not check_issn(isbn):
                field_values["issn"].append(isbn)
            if not check_ean(isbn):
                field_values["ean"].append(isbn)

    for k, values in field_values.items():
        with open(os.path.join(reports_directory, f"calculated_field_{k}.yml"), "w", encoding="utf-8") as outfile:
            if values and isinstance(values[0], list):
                yaml.dump(values, outfile, default_flow_style=False, allow_unicode=True)
            else:
                yaml.dump(
                    list(sorted(set(values), key=lambda x: romanize(x))),
                    outfile,
                    default_flow_style=False,
                    allow_unicode=True,
                )


def test_report_dewey(reports_directory: str):
    weird_dewey: list[dict[int, str]] = list()
    for entry in all_entries():
        dewey = dewey_from_a04_a05(entry[4], entry[5])
        if dewey:
            if not is_valid_dewey_strict(dewey):
                weird_dewey.append(minimal_entry(entry, [0, 1, 2, 4, 5, 6]))
        else:
            if none_if_empty_or_stripped(entry[4]):
                weird_dewey.append(minimal_entry(entry, [0, 1, 2, 4, 5, 6]))
    with open(os.path.join(reports_directory, f"weird_dewey.yml"), "w", encoding="utf-8") as outfile:
        yaml.dump(weird_dewey, outfile, default_flow_style=False, allow_unicode=True)


def test_report_entry_numbers(reports_directory: str):
    all_entry_numbers: dict[str, dict[int, str]] = dict()
    no_entry_numbers: list[dict[int, str]] = list()
    non_numeric: defaultdict[str, list[dict[int, str]]] = defaultdict(list)
    duplicate_entry_numbers: defaultdict[str, list[dict[int, str]]] = defaultdict(list)
    for entry in all_entries():
        entry_numbers = entry_numbers_from_a04_a05_a06_a07_a08_a18_a19(
            entry[4], entry[5], entry[6], entry[7], entry[8], entry[18], entry[19]
        )
        if not entry_numbers:
            no_entry_numbers.append(minimal_entry(entry, [0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 17, 18, 19]))
        else:
            for n in entry_numbers:
                if n in all_entry_numbers:
                    if not duplicate_entry_numbers[n]:
                        duplicate_entry_numbers[n].append(all_entry_numbers[n])
                    duplicate_entry_numbers[n].append(
                        minimal_entry(entry, [0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 17, 18, 19])
                    )
                else:
                    all_entry_numbers[n] = minimal_entry(entry, [0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 17, 18, 19])
                if not n.isnumeric():
                    non_numeric[n].append(minimal_entry(entry, [0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 17, 18, 19]))
    with open(os.path.join(reports_directory, f"entry_numbers_no_entry_numbers.yml"), "w", encoding="utf-8") as outfile:
        yaml.dump(no_entry_numbers, outfile, default_flow_style=False, allow_unicode=True)
    with open(os.path.join(reports_directory, f"entry_numbers_non_numeric.yml"), "w", encoding="utf-8") as outfile:
        yaml.dump(dict(non_numeric), outfile, default_flow_style=False, allow_unicode=True)
    with open(os.path.join(reports_directory, f"entry_numbers_duplicate.yml"), "w", encoding="utf-8") as outfile:
        yaml.dump(dict(duplicate_entry_numbers), outfile, default_flow_style=False, allow_unicode=True)


def test_report_weird_names(reports_directory: str):
    weird_translators: list[str] = []
    weird_authors: list[str] = []
    weird_curators: list[str] = []
    weird_donors: list[str] = []
    valid_name_re = re.compile(r"[A-ZΑ-Ω\-]+,[A-ZΑ-Ω\.]*\.?")
    for entry in all_entries():
        translator = translator_from_a06(entry[6])
        if translator:
            translators = translator.split("!!")
            for translator in translators:
                if not valid_name_re.fullmatch(translator):
                    weird_translators.append(translator)
        authors = authors_from_a01(entry[1])
        for author in authors:
            if not valid_name_re.fullmatch(author):
                weird_authors.append(author)
        curator = curator_from_a16(entry[16])
        if curator:
            curators = curator.split("!!")
            for curator in curators:
                if not valid_name_re.fullmatch(curator):
                    weird_curators.append(curator)
        donor = donation_from_a17_a30(entry[17], entry[30])
        if donor:
            donors = donor.split("!!")
            for donor in donors:
                if not valid_name_re.fullmatch(donor):
                    weird_donors.append(donor)
    for field, thelist in [
        ("translators", weird_translators),
        ("authors", weird_authors),
        ("curators", weird_curators),
        ("donors", weird_donors),
    ]:
        with open(os.path.join(reports_directory, f"weird_{field}.yml"), "w", encoding="utf-8") as outfile:
            yaml.dump(sorted(set(thelist)), outfile, default_flow_style=False, allow_unicode=True)


def test_report_bools(reports_directory: str):
    entries_with_cd: list[dict[int, str]] = list()
    entries_with_dvd: list[dict[int, str]] = list()
    entries_with_offprint: list[dict[int, str]] = list()
    for entry in all_entries():
        if has_cd_from_a02_a03_a12_a13_a14_a17_a18_a22_a30(
            [
                entry[2],
                entry[3],
                entry[12],
                entry[13],
                entry[14],
                entry[17],
                entry[18],
                entry[22],
                entry[30],
            ]
        ):
            entries_with_cd.append(minimal_entry(entry, list(range(0, 31))))
        if has_dvd_from_a30(
            [
                entry[30],
            ]
        ):
            entries_with_dvd.append(minimal_entry(entry, list(range(0, 31))))
        if offprint_from_a17_a21_a30(entry[17], entry[21], entry[30]):
            entries_with_offprint.append(minimal_entry(entry, list(range(0, 31))))
    with open(os.path.join(reports_directory, f"entries_with_cd.yml"), "w", encoding="utf-8") as outfile:
        yaml.dump(entries_with_cd, outfile, default_flow_style=False, allow_unicode=True)
    with open(os.path.join(reports_directory, f"entries_with_dvd.yml"), "w", encoding="utf-8") as outfile:
        yaml.dump(entries_with_dvd, outfile, default_flow_style=False, allow_unicode=True)
    with open(os.path.join(reports_directory, f"entries_with_offprint.yml"), "w", encoding="utf-8") as outfile:
        yaml.dump(entries_with_offprint, outfile, default_flow_style=False, allow_unicode=True)


def test_report_donors(reports_directory: str):
    count_map: defaultdict[str, int] = defaultdict(int)
    for entry in all_entries():
        donors = donation_from_a17_a30(entry[17], entry[30])
        if donors:
            for donor in donors.split("!!"):
                count_map[donor] = count_map[donor] + 1

    with open(os.path.join(reports_directory, f"donors_count.yml"), "w", encoding="utf-8") as outfile:
        yaml.dump(
            dict(sorted(count_map.items(), key=lambda x: x[1], reverse=True)),
            outfile,
            default_flow_style=False,
            allow_unicode=True,
        )


def test_report_isbns(reports_directory: str):
    entries_with_errors: dict[str, dict[str, str | list[Optional[str]] | dict[int, str]]] = dict()
    for entry in all_entries():
        result = isbn_from_a17_a18_a19_a22_a30(
            entry[17],
            entry[18],
            entry[19],
            entry[22],
            entry[30],
        )
        if not result:
            continue
        checks = [check_isbn(result), check_issn(result), check_ean(result)]
        if None in checks:
            continue
        else:
            entries_with_errors[entry[0]] = {
                "entry": minimal_entry(entry, list(range(0, 31))),
                "isbn": result,
                "errors": checks,
            }
    with open(
        os.path.join(reports_directory, f"entries_with_invalid_isbn_issn_ean.yml"), "w", encoding="utf-8"
    ) as outfile:
        yaml.dump(entries_with_errors, outfile, default_flow_style=False, allow_unicode=True)


def test_report_extracted_fields(reports_directory: str):
    converted_entries: list[dict[str, Optional[str] | int | bool | list[str] | dict[int, str]]] = []
    for entry in all_entries():
        converted_entry: dict[str, Optional[str] | int | bool | list[str] | dict[int, str]] = dict()
        converted_entry["dbase_number"] = entry[0]
        converted_entry["authors"] = authors_from_a01(entry[1])

        language = language_from_a01(entry[1])
        if language:
            converted_entry["language"] = language

        title = title_from_a02(entry[2])
        if title:
            converted_entry["title"] = title

        subtitle = subtitle_from_a03(entry[3])
        if subtitle:
            converted_entry["subtitle"] = subtitle

        dewey = dewey_from_a04_a05(entry[4], entry[5])
        if dewey:
            converted_entry["dewey"] = dewey

        converted_entry["entry_numbers"] = entry_numbers_from_a04_a05_a06_a07_a08_a18_a19(
            entry[4], entry[5], entry[6], entry[7], entry[8], entry[18], entry[19]
        )

        translator = translator_from_a06(entry[6])
        if translator:
            converted_entry["translators"] = translator.split("!!")
            # for single_translator in translators:
            #     translator_surname_name = single_translator.split(",", maxsplit=1)
            #     if len(translator_surname_name) == 2:
            #         field_values["translator_family_name"].append(translator_surname_name[0])
            #         if translator_surname_name[1].endswith("."):
            #             field_values["translator_name_abbreviations"].append(translator_surname_name[1])
            #         else:
            #             field_values["translator_names"].append(translator_surname_name[1])

        edition = edition_from_a07(entry[7])
        if edition:
            converted_entry["edition"] = edition

        editor = editor_from_a08_a09(entry[8], entry[9])
        if editor:
            converted_entry["editor"] = f"{editor[0]} // {editor[1]}"

        edition_year = edition_year_from_a09_a10(entry[9], entry[10])
        if edition_year:
            converted_entry["edition_year"] = edition_year

        pages = pages_from_a11(entry[11])
        if pages:
            converted_entry["pages"] = pages

        topic_list = topics_from_a12_to_a15_a20_a22_to_a24(
            [
                entry[12],
                entry[13],
                entry[14],
                entry[15],
                entry[20],
                entry[22],
                entry[23],
                entry[24],
            ]
        )
        converted_entry["topics"] = topic_list

        curator = curator_from_a16(entry[16])
        if curator:
            converted_entry["curator"] = curator

        copies = copies_from_a17_a18_a30(entry[17], entry[18], entry[30])
        if copies:
            converted_entry["copies"] = copies

        donation = donation_from_a17_a30(entry[17], entry[30])
        if donation:
            converted_entry["donors"] = donation.split("!!")

        volume = volume_from_a17_a18_a20_a30(entry[17], entry[18], entry[20], entry[30])
        if volume:
            converted_entry["volume"] = volume

        material = material_from_a18_a30(entry[18], entry[30])
        if material:
            converted_entry["material"] = material

        notes = notes_from_a17_a18_a21_a30(entry[17], entry[18], entry[21], entry[30])
        if notes:
            converted_entry["notes"] = notes

        converted_entry["has_cd"] = has_cd_from_a02_a03_a12_a13_a14_a17_a18_a22_a30(
            [
                entry[2],
                entry[3],
                entry[12],
                entry[13],
                entry[14],
                entry[17],
                entry[18],
                entry[22],
                entry[30],
            ]
        )

        converted_entry["has_dvd"] = has_dvd_from_a30(
            [
                entry[30],
            ]
        )

        converted_entry["offprint"] = offprint_from_a17_a21_a30(entry[17], entry[21], entry[30])

        isbn = isbn_from_a17_a18_a19_a22_a30(entry[17], entry[18], entry[19], entry[22], entry[30])
        if isbn:
            if not check_isbn(isbn):
                converted_entry["isbn"] = isbn
            if not check_issn(isbn):
                converted_entry["issn"] = isbn
            if not check_ean(isbn):
                converted_entry["ean"] = isbn

        converted_entry["original_entry"] = entry
        converted_entries.append(converted_entry)

    with open(
        os.path.join(
            os.path.dirname(__file__), os.pardir, "src", "skoufas_dbf_reader", "data", "converted_entries.yml"
        ),
        "w",
        encoding="utf-8",
    ) as outfile:
        yaml.dump(
            {"converted_entries": converted_entries},
            outfile,
            default_flow_style=False,
            allow_unicode=True,
            sort_keys=False,
        )



""" Utilities for conversions """
import os
import re
from functools import cache
from typing import Any, Optional

import yaml


def read_yaml_data(code: str) -> Any:
    """Return the only object from a yaml file in the data directory"""
    with open(os.path.join(os.path.dirname(__file__), "data", f"{code}.yml"), "r", encoding="utf-8") as stream:
        parsed_yaml = yaml.safe_load(stream)
        return parsed_yaml[code]


def none_if_empty_or_stripped(i: Optional[str]) -> Optional[str]:
    """Cleanup"""
    if not i:
        return None
    if not i.strip():
        return None
    return i.strip()


@cache
def all_entries() -> list[dict[int, str]]:
    """All entries converted from a DBF file"""
    data = read_yaml_data("entries")
    for entry in data:
        for i in range(0, 31):
            if i not in entry:
                entry[i] = None
    return data


def romanize(greek_text: Optional[str]) -> str:
    """Return the ISO 843:1997 transcription of the input Greek text.
    Any non-Greek characters will be ignored and printed as they were.

    This function is (c) George Schizas, released under the Apache 2,0 licence
    See https://github.com/gschizas/RomanizePython/blob/master/src/romanize/romanize.py
    """

    if not greek_text:
        return ""
    result = ""
    cursor = 0
    while cursor < len(greek_text):
        letter = greek_text[cursor]
        prev_letter = greek_text[cursor - 1] if cursor > 0 else ""
        next_letter = greek_text[cursor + 1] if cursor < len(greek_text) - 1 else ""
        third_letter = greek_text[cursor + 2] if cursor < len(greek_text) - 2 else ""

        is_upper = letter.upper() == letter
        is_upper_next = next_letter.upper() == next_letter
        letter = letter.lower()
        prev_letter = prev_letter.lower()
        next_letter = next_letter.lower()
        third_letter = third_letter.lower()

        simple_translation_greek = "άβδέζήιίϊΐκλνξόπρσςτυύϋΰφωώ"
        simple_translation_latin = "avdeziiiiiklnxoprsstyyyyfoo"

        digraph_translation_greek = "θχψ"
        digraph_translation_latin = "thchps"

        digraph_ypsilon_greek = "αεη"
        digraph_ypsilon_latin = "aei"
        digraph_ypsilon_beta = "βγδζλμνραάεέηήιίϊΐοόυύϋΰωώ"
        digraph_ypsilon_phi = "θκξπστφχψ"

        if letter in simple_translation_greek:
            new_letter = simple_translation_latin[simple_translation_greek.index(letter)]
        elif letter in digraph_translation_greek:
            diphthong_index = digraph_translation_greek.index(letter)
            new_letter = digraph_translation_latin[diphthong_index * 2 : diphthong_index * 2 + 2]
        elif letter in digraph_ypsilon_greek:
            new_letter = digraph_ypsilon_latin[digraph_ypsilon_greek.index(letter)]
            if next_letter in ["υ", "ύ"]:
                if third_letter in digraph_ypsilon_beta:
                    new_letter += "v"
                    cursor += 1
                elif third_letter in digraph_ypsilon_phi:
                    new_letter += "f"
                    cursor += 1
        elif letter == "γ":
            if next_letter == "γ":
                new_letter = "ng"
                cursor += 1
            elif next_letter == "ξ":
                new_letter = "nx"
                cursor += 1
            elif next_letter in "χ":
                new_letter = "nch"
                cursor += 1
            else:
                new_letter = "g"
        elif letter == "μ":
            if next_letter == "π":
                if prev_letter.strip() == "" or third_letter.strip() == "":
                    new_letter = "b"
                    cursor += 1
                else:
                    new_letter = "mp"
                    cursor += 1
            else:
                new_letter = "m"
        elif letter == "ο":
            new_letter = "o"
            if next_letter in ["υ", "ύ"]:
                new_letter += "u"
                cursor += 1
        else:
            new_letter = letter
        if is_upper:
            new_letter = new_letter[0].upper() + (new_letter[1:].upper() if is_upper_next else new_letter[1:].lower())
        result += new_letter
        cursor += 1
    return result


def check_isbn(isbn: str) -> Optional[str]:
    isbn = isbn.replace("-", "").replace(" ", "").upper()
    match = re.search(r"^(\d{9})(\d|X)$", isbn)
    if not match:
        return f"Invalid isbn format (len {len(isbn)})"

    digits = match.group(1)
    check_digit = 10 if match.group(2) == "X" else int(match.group(2))

    result = sum((i + 1) * int(digit) for i, digit in enumerate(digits))
    if (result % 11) == check_digit:
        return None
    else:
        return f"Invalid check code {result % 11} != {check_digit}"


def check_issn(issn: str) -> Optional[str]:
    issn = issn.replace("-", "").replace(" ", "").upper()
    match = re.search(r"^(\d{7})(\d|X)$", issn)
    if not match:
        return f"Invalid issn format (len {len(issn)})"

    digits = match.group(1)
    check_digit = 10 if match.group(2) == "X" else int(match.group(2))
    result = sum((i + 1) * int(digit) for i, digit in enumerate(digits))
    if (result % 11) == check_digit:
        return None
    else:
        return f"Invalid check code {result % 11} != {check_digit}"


def check_ean(ean: str) -> Optional[str]:
    ean = ean.replace("-", "").replace(" ", "").upper()
    match = re.search(r"^(\d+)(\d)$", ean)
    if not match:
        return f"Invalid ean format"
    if len(ean) not in (14, 13, 12, 8):
        return f"Invalid ean format (len {len(ean)})"

    digits = match.group(1)
    check_digit = match.group(2)
    result = str((10 - sum((3, 1)[i % 2] * int(n) for i, n in enumerate(reversed(digits)))) % 10)
    if result == check_digit:
        return None
    else:
        return f"Invalid check code {result} != {check_digit}"


def is_valid_dewey_strict(d: str):
    strict_dewey_res = [
        re.compile(r"[0-9]{3}\.[0-9]+ [^0-9]+"),
        re.compile(r"[0-9]{3}\.[0-9]+"),
        re.compile(r"[0-9]{3} [^0-9]+"),
        re.compile(r"[0-9]{3}"),
    ]
    for dewey_re in strict_dewey_res:
        if dewey_re.fullmatch(d):
            return True
    return False



"""Convert dbf files to human readable yaml"""
from __future__ import annotations

import sys
from collections import OrderedDict

import dbfread
import yaml


def convert_dbf_to_yaml(from_dbf_file: str, to_yaml_file: str):
    """Convert dbf files to human readable yaml"""
    entries: list[dict[int, str | int]] = list()
    with dbfread.DBF(filename=from_dbf_file, encoding="CP737") as dbf:
        count: int
        record: dict[str, str]
        for count, record in enumerate(dbf):
            entry: dict[int, str | int] = dict()
            entry[0] = count + 1
            for name, value in record.items():
                if not value:
                    continue
                idx = int(name.replace("A", ""))
                entry[idx] = value
            entries.append(entry)

    data = {"entries": entries}
    with open(to_yaml_file, "w", encoding="utf-8") as outfile:
        yaml.dump(data, outfile, default_flow_style=False, allow_unicode=True)


def main():
    """Convert the dbf file passed to the first argument to a yaml file in the file passed as the second argument"""
    convert_dbf_to_yaml(sys.argv[1], sys.argv[2])


if __name__ == "__main__":
    main()



""" Functions that extract information given specific strings """
from collections import OrderedDict
from typing import Optional

from skoufas_dbf_reader.correction_data import *
from skoufas_dbf_reader.utilities import none_if_empty_or_stripped


def has_language(a01: Optional[str]) -> bool:
    """Check values for language at the end"""
    if not a01:
        return False
    for code in language_codes().keys():
        if a01.endswith(f" {code}"):
            return True
    return False


def authors_from_a01(a01: Optional[str]) -> list[str]:
    """Get author from A01 DBF record"""
    if not a01:
        return []
    if not has_author(a01):
        return []
    strip_finals = list(language_codes().keys()) + [
        " Ι",
        " .",
    ]
    for final in strip_finals:
        if a01.endswith(" " + final):
            a01 = a01.replace(final, "")
    author = a01.strip()
    author = author_corrections().get(author, author)
    if not author:
        author = ""
    return author.split("!!")


def language_from_a01(a01: Optional[str]) -> Optional[str]:
    """Check values for language at the end"""
    if not a01:
        return None
    if has_language(a01):
        for language, isolanguage in language_codes().items():
            if a01.endswith(language):
                return isolanguage
    return None


def title_from_a02(a02: Optional[str]) -> Optional[str]:
    """Cleanup"""
    return none_if_empty_or_stripped(a02)


def subtitle_from_a03(a03: Optional[str]) -> Optional[str]:
    """Cleanup"""
    return none_if_empty_or_stripped(a03)


def dewey_from_a04_a05(a04: Optional[str], a05: Optional[str]) -> Optional[str]:
    """Cleanup and replace known issues"""
    value4 = none_if_empty_or_stripped(a04)
    if value4:
        correction = field04_corrections().get(value4, value4)
        if correction:
            if isinstance(correction, str):
                value4 = none_if_empty_or_stripped(correction)
            else:
                value4 = None
        else:
            value4 = None

    if value4:
        value4 = value4.replace("Χ. Σ.", "ΧΣ")
        value4 = value4.replace("Χ. Σ", "ΧΣ")

        value4 = value4.replace("X.S.", "ΧΣ")
        value4 = value4.replace("Χ.Σ.", "ΧΣ")

        value4 = value4.replace("Χ.Σ", "ΧΣ")
        value4 = value4.replace("X.S", "ΧΣ")
        value4 = value4.replace("X.Σ", "ΧΣ")
        value4 = value4.replace("Σ.Σ", "ΧΣ")

        value4 = value4.replace("Χ.Χ.", "ΧΧ")
        value4 = value4.replace("Χ.Χ", "ΧΧ")

    if value4:
        for dewey_re in dewey_re1:
            dewey_match = dewey_re.fullmatch(value4)
            if dewey_match:
                return f"{dewey_match[1]}".strip()
        for dewey_re in dewey_re2:
            dewey_match = dewey_re.fullmatch(value4)
            if dewey_match:
                return f"{dewey_match[1]} {dewey_match[2]}".strip()

    value5 = none_if_empty_or_stripped(a05)
    if value5:
        correction = field05_corrections().get(value5, value5)
        if correction:
            if isinstance(correction, dict):
                value5 = correction.get("dewey", None)
            else:
                value5 = None
        else:
            value5 = None
    if value5 and isinstance(value5, str):
        for dewey_re in dewey_re1:
            dewey_match = dewey_re.fullmatch(value5)
            if dewey_match:
                return f"{dewey_match[1]}".strip()
        for dewey_re in dewey_re2:
            dewey_match = dewey_re.fullmatch(value5)
            if dewey_match:
                return f"{dewey_match[1]} {dewey_match[2]}".strip()

    return None


def entry_numbers_from_a04_a05_a06_a07_a08_a18_a19(
    a04: Optional[str],
    a05: Optional[str],
    a06: Optional[str],
    a07: Optional[str],
    a08: Optional[str],
    a18: Optional[str],
    a19: Optional[str],
) -> list[str]:
    """Cleanup, read additional numbers from a06"""

    def cleanup_single_value(
        field_name: str,
        original_value: Optional[str],
        corrections: dict[str, Optional[str | dict[str, str] | dict[str, str | bool]]],
        ignore_if_not_in_correction: bool,
    ) -> str:
        output = none_if_empty_or_stripped(original_value)
        if output:
            if output in corrections:
                correction = corrections[output]
                if correction is None:
                    return ""
                elif isinstance(correction, str):
                    if ignore_if_not_in_correction:
                        return ""
                    else:
                        return "-" + str(correction)
                else:
                    output = dict(correction).get("series", "")
                    if not isinstance(output, str):
                        raise Exception(f"Invalid correction for field {field_name} [{original_value}]")
                    if correction.get("use_dash", True):
                        return "-" + output
                    else:
                        return output
            else:
                if ignore_if_not_in_correction:
                    return ""
                else:
                    return output
        return ""

    value4 = cleanup_single_value("A04", a04, field04_corrections(), True) + "-"  # type: ignore
    value5 = cleanup_single_value("A05", a05, field05_corrections(), False)  # type: ignore
    value6 = cleanup_single_value("A06", a06, field06_corrections(), True)  # type: ignore
    value7 = cleanup_single_value("A07", a07, field07_corrections(), True)  # type: ignore
    value8 = cleanup_single_value("A08", a08, field08_corrections(), True)  # type: ignore
    value18 = cleanup_single_value("A18", a18, field18_corrections(), True)  # type: ignore
    value19 = cleanup_single_value("A19", a19, field19_corrections(), True)  # type: ignore

    value = value4 + value5 + value6 + value7 + value8 + value18 + value19
    # Use a dict to remove duplicates
    entries_dict: dict[str, None] = dict()
    for v in value.split("-"):
        if v.strip():
            entries_dict[v.strip()] = None
    return list(entries_dict)


def translator_from_a06(a06: Optional[str]) -> Optional[str]:
    """Cleanup, replace special cases"""
    value = none_if_empty_or_stripped(a06)
    if not value:
        return None
    if value in field06_corrections():
        correction = field06_corrections()[value]
        if correction is None:
            return None
        elif isinstance(correction, str):
            return None
        else:
            value = dict(correction).get("translator", None)
            if not value:
                return None
            if not isinstance(value, str):
                raise Exception(f"Invalid correction for field A06 [{a06}]")
    value = translator_corrections().get(value, value)
    if not value:
        return None
    return value


def edition_from_a07(a07: Optional[str]) -> Optional[str]:
    """Cleanup, replace special cases"""
    value = none_if_empty_or_stripped(a07)
    if not value:
        return None
    if value in field07_corrections():
        value = field07_corrections().get(value)
        if not isinstance(value, str):
            return None
    if not value:
        return None
    return value


def editor_from_a08_a09(a08: Optional[str], a09: Optional[str]) -> Optional[tuple[Optional[str], Optional[str]]]:
    """Cleanup, replace special cases"""
    a08 = none_if_empty_or_stripped(a08)
    if not a08:
        a08 = None
    else:
        if a08 in field08_corrections():
            correction = field08_corrections().get(a08)
            if not correction:
                a08 = None
            elif isinstance(correction, dict):
                a08 = correction.get("editor", None)
            else:
                a08 = correction
    a09 = none_if_empty_or_stripped(a09)
    if not a09:
        a09 = None
    else:
        if a09 in field09_corrections():
            correction = field09_corrections().get(a09)
            if not correction:
                a09 = None
            elif isinstance(correction, dict):
                a09 = correction.get("place", None)
            else:
                a09 = correction
    if not a08 and not a09:
        return None
    correction = editor_corrections().get(f"{a08} // {a09}")
    if correction:
        correction = correction.split(" // ")
        a08 = correction[0]
        a09 = correction[1]
    return (a08, a09)


def edition_year_from_a09_a10(a09: Optional[str], a10: Optional[str]) -> Optional[int]:
    """Cleanup,handle special cases"""
    a09 = none_if_empty_or_stripped(a09)
    if a09 and a09 in field09_corrections():
        correction = field09_corrections().get(a09)
        if correction and isinstance(correction, dict) and "year" in correction:
            return int(correction["year"])

    a10 = none_if_empty_or_stripped(a10)
    if not a10:
        return None
    corrected = field10_corrections().get(a10, a10)
    if not corrected:
        return None
    if not corrected.isnumeric():
        raise Exception(f"invalid date [{a10}]")
    return int(corrected)


def pages_from_a11(a11: Optional[str]) -> Optional[int]:
    """Cleanup, return an int from various expressions"""
    a11 = none_if_empty_or_stripped(a11)
    if not a11:
        return None
    corrected = field11_corrections().get(a11, a11)
    if not corrected:
        return None
    pages_match = valid_pages_re.fullmatch(corrected)
    if not pages_match:
        raise Exception(f"invalid pages [{a11}]")
    return int(pages_match.group(1))


def topics_from_a12_to_a15_a20_a22_to_a24(
    many_lines: Optional[list[Optional[str]]],
) -> list[str]:
    """Cleanup, make unique, handle special cases"""
    if not many_lines:
        return []
    topics: OrderedDict[str, None] = OrderedDict()
    for line in many_lines:
        line = none_if_empty_or_stripped(line)
        if line:
            match = topic_in_paren_re.fullmatch(line)
            if match:
                in_parenthesis = match[1].strip()
                in_parenthesis = topic_replacements().get(in_parenthesis, in_parenthesis)
                if in_parenthesis and in_parenthesis.strip():
                    topics[in_parenthesis.strip()] = None
                # remove match and parenthesis
                line = line.replace(match[1], "").replace("()", "")
            for topic in line.split("-"):
                if topic:
                    topic = topic_replacements().get(topic, topic)
                    if topic:
                        topics[topic.strip()] = None
    return list(topics.keys())


def curator_from_a16(a16: Optional[str]) -> Optional[str]:
    """Cleanup"""
    a16 = none_if_empty_or_stripped(a16)
    if not a16:
        return None
    if a16 in field16_corrections():
        return field16_corrections()[a16]
    else:
        return a16


def has_cd_from_a02_a03_a12_a13_a14_a17_a18_a22_a30(many_lines: Optional[list[Optional[str]]]) -> bool:
    """Look for CD in the lines passed"""
    if not many_lines:
        return False
    for line in many_lines:
        if line and has_cd_re.search(line):
            return True
    return False


def has_dvd_from_a30(many_lines: Optional[list[Optional[str]]]) -> bool:
    """Look for DVD in the lines passed"""
    if not many_lines:
        return False
    for line in many_lines:
        if line and has_dvd_re.search(line):
            return True
    return False


def copies_from_a17_a18_a30(a17: Optional[str], a18: Optional[str], a30: Optional[str]) -> Optional[int]:
    """Use corrections to look for number of copies"""
    a17 = none_if_empty_or_stripped(a17)
    if a17:
        correction = field17_corrections().get(a17)
        if correction and isinstance(correction, dict):
            copies = correction.get("copies")
            if copies:
                if isinstance(copies, int):
                    return copies
                else:
                    raise Exception(f"Invalid correction for A17 [{a17}]")
    a18 = none_if_empty_or_stripped(a18)
    if a18:
        correction = field18_corrections().get(a18)
        if correction and isinstance(correction, dict):
            copies = correction.get("copies")
            if copies:
                if isinstance(copies, int):
                    return copies
                else:
                    raise Exception(f"Invalid correction for A17 [{a18}]")
    a30 = none_if_empty_or_stripped(a30)
    if a30:
        correction = field30_corrections().get(a30)
        if correction and isinstance(correction, dict):
            copies = correction.get("copies")
            if copies:
                if isinstance(copies, int):
                    return copies
                else:
                    raise Exception(f"Invalid correction for A30 [{a30}]")
    return None


def donation_from_a17_a30(a17: Optional[str], a30: Optional[str]) -> Optional[str]:
    """Use corrections to look for donations"""
    a17 = none_if_empty_or_stripped(a17)
    if a17:
        correction = field17_corrections().get(a17)
        if correction and isinstance(correction, dict):
            donation = correction.get("donation")
            if donation:
                if isinstance(donation, str):
                    return donation
                else:
                    raise Exception(f"Invalid correction for A17 [{a17}]")
    a30 = none_if_empty_or_stripped(a30)
    if a30:
        correction = field30_corrections().get(a30)
        if correction and isinstance(correction, dict):
            donation = correction.get("donation")
            if donation:
                if isinstance(donation, str):
                    return donation
                else:
                    raise Exception(f"Invalid correction for A30 [{a30}]")
    return None


def offprint_from_a17_a21_a30(a17: Optional[str], a21: Optional[str], a30: Optional[str]) -> bool:
    """Use corrections to look for offprint or the word ΑΝΑΤΥΠΟ"""
    a17 = none_if_empty_or_stripped(a17)
    if a17:
        correction = field17_corrections().get(a17)
        if correction and isinstance(correction, dict):
            offprint = correction.get("offprint")
            if offprint:
                if isinstance(offprint, bool):
                    return offprint
                else:
                    raise Exception(f"Invalid correction for A17 [{a17}]")
        if "ΑΝΑΤΥΠΟ" in a17:
            return True
    if a21 and "ΑΝΑΤΥΠΟ" in a21:
        return True
    a30 = none_if_empty_or_stripped(a30)
    if a30:
        correction = field30_corrections().get(a30)
        if correction and isinstance(correction, dict):
            offprint = correction.get("offprint")
            if offprint:
                if isinstance(offprint, bool):
                    return offprint
                else:
                    raise Exception(f"Invalid correction for A30 [{a30}]")
        elif "ΑΝΑΤΥΠΟ" in a30:
            return True
    return False


def volume_from_a17_a18_a20_a30(
    a17: Optional[str], a18: Optional[str], a20: Optional[str], a30: Optional[str]
) -> Optional[str]:
    """Use corrections to look for volume"""

    def read_from_single_field(
        field_name: str,
        value: Optional[str],
        corrections: dict[str, str | dict[str, str | bool | int] | None],
        current_result: str,
    ) -> str:
        value = none_if_empty_or_stripped(value)
        if value:
            correction = corrections.get(value)
            if correction and isinstance(correction, dict):
                result = correction.get("volume")
                if result:
                    if isinstance(result, str):
                        if current_result:
                            return current_result + "\n" + result
                        else:
                            return result
                    else:
                        raise Exception(f"Invalid correction for {field_name} [{value}]")
        return current_result

    result = read_from_single_field("A17", a17, field17_corrections(), "")
    result = read_from_single_field("A18", a18, field18_corrections(), result)
    result = read_from_single_field("A20", a20, field20_corrections(), result)
    result = read_from_single_field("A30", a30, field30_corrections(), result)
    return none_if_empty_or_stripped(result)


def material_from_a18_a30(a18: Optional[str], a30: Optional[str]) -> Optional[str]:
    """Use corrections to look for volume"""

    def read_from_single_field(
        field_name: str,
        value: Optional[str],
        corrections: dict[str, str | dict[str, str | bool | int] | None],
        current_result: str,
    ) -> str:
        value = none_if_empty_or_stripped(value)
        if value:
            correction = corrections.get(value)
            if correction and isinstance(correction, dict):
                result = correction.get("material")
                if result:
                    if isinstance(result, str):
                        if current_result:
                            return current_result + "\n" + result
                        else:
                            return result
                    else:
                        raise Exception(f"Invalid correction for {field_name} [{value}]")
        return current_result

    result = read_from_single_field("A18", a18, field18_corrections(), "")
    result = read_from_single_field("A30", a30, field30_corrections(), result)
    return none_if_empty_or_stripped(result)


def notes_from_a17_a18_a21_a30(
    a17: Optional[str], a18: Optional[str], a21: Optional[str], a30: Optional[str]
) -> Optional[str]:
    """Read from all three fields, apply corrections"""

    def read_from_single_field(
        field_name: str,
        value: Optional[str],
        corrections: dict[str, str | dict[str, str | bool | int] | None],
        current_result: str,
    ) -> str:
        result = None
        value = none_if_empty_or_stripped(value)
        if not value:
            result = None
        elif value in corrections:
            correction = corrections.get(value)
            if correction:
                if isinstance(correction, dict):
                    notes_in_dict = correction.get("notes")
                    if not notes_in_dict:
                        result = None
                    elif isinstance(notes_in_dict, str):
                        result = notes_in_dict
                    else:
                        raise Exception(f"Invalid correction for {field_name} [{value}]")
                if isinstance(correction, str):
                    result = correction
            else:
                result = None
        else:
            result = value

        if result:
            if current_result:
                return current_result + "\n" + result
            else:
                return result
        else:
            return current_result

    result = read_from_single_field("A17", a17, field17_corrections(), "")

    a18 = none_if_empty_or_stripped(a18)
    if a18 and a18 in field18_corrections():
        correction = field18_corrections().get(a18)
        if correction:
            if isinstance(correction, dict):
                notes_in_dict = correction.get("notes")
                if notes_in_dict and isinstance(notes_in_dict, str):
                    if result:
                        result += "\n" + notes_in_dict
                    else:
                        result = notes_in_dict

    result = read_from_single_field("A21", a21, {}, result)
    result = read_from_single_field("A30", a30, field30_corrections(), result)

    return none_if_empty_or_stripped(result)


def isbn_from_a17_a18_a19_a22_a30(
    a17in: Optional[str], a18in: Optional[str], a19in: Optional[str], a22in: Optional[str], a30in: Optional[str]
) -> Optional[str]:
    """Cleanup"""

    # Only use corrections from a17
    a17 = none_if_empty_or_stripped(a17in)
    if not a17:
        a17 = ""
    if a17 in field17_corrections():
        correction = field17_corrections()[a17]

        if correction:
            if isinstance(correction, dict):
                if "isbn" in correction:
                    if correction and isinstance(correction["isbn"], str):
                        a17 = correction["isbn"]
                    else:
                        a17 = ""
                else:
                    a17 = ""
            else:
                a17 = ""
        else:
            a17 = ""
    else:
        a17 = ""

    # Use a18 unless there's a correction
    a18 = none_if_empty_or_stripped(a18in)
    if a18:
        if a18 in field18_corrections():
            a18 = ""
    else:
        a18 = ""

    a19 = none_if_empty_or_stripped(a19in)
    if a19:
        if a19 in field19_corrections():
            a19 = ""
    else:
        a19 = ""

    a22 = none_if_empty_or_stripped(a22in)
    if not a22:
        a22 = ""
    if not a22_has_isbn_part_re.fullmatch(a22):
        a22 = ""

    # Only use corrections from a30
    a30 = none_if_empty_or_stripped(a30in)
    if not a30:
        a30 = ""
    if a30 in field30_corrections():
        correction = field30_corrections()[a30]
        if correction:
            if isinstance(correction, dict):
                if "isbn" in correction:
                    if correction and isinstance(correction["isbn"], str):
                        a30 = correction["isbn"]
                    else:
                        a30 = ""
                else:
                    a30 = ""
            else:
                a30 = ""
        else:
            a30 = ""
    else:
        a30 = ""

    result = (a17 + a18 + a19 + a22 + a30).replace(" ", "").replace(".", "")
    result = none_if_empty_or_stripped(result)
    return result



from __future__ import annotations

import os
import pprint
import shutil
import sys
from collections import defaultdict

import yaml
from snakemd import Document, Inline, MDList, Table

from skoufas_dbf_reader.correction_data import plain_author_re
from skoufas_dbf_reader.field_extractors import *
from skoufas_dbf_reader.utilities import all_entries, check_ean, check_isbn, check_issn, is_valid_dewey_strict, romanize


def report_single_fields(reports_directory: str):
    field_values: list[set[str]] = [set() for _ in range(31)]
    os.makedirs(os.path.join(reports_directory, "single-field"), exist_ok=True)
    for entry in all_entries():
        for i in range(1, 31):
            if i in entry and entry[i] and entry[i].strip():
                field_values[i].add(entry[i].strip())
    for i in range(1, 31):
        doc = Document()
        doc.add_heading(f"Τιμές στη θέση {i:02}")
        doc.add_unordered_list([f"`{v}`" for v in sorted(field_values[i])])
        with open(
            os.path.join(reports_directory, "single-field", f"field_{i:02}.md"), "w", encoding="utf-8"
        ) as outfile:
            outfile.write(str(doc))

    doc = Document()
    doc.add_heading("Τιμές στις στήλες των καρτελών")
    doc.add_unordered_list([str(Inline(f"Στήλη {i}", link=f"./field_{i:02}.html")) for i in range(1, 31)])
    with open(os.path.join(reports_directory, "single-field", "index.md"), "w", encoding="utf-8") as outfile:
        outfile.write(str(doc))


def report_single_extracted_fields(reports_directory: str):
    field_values: defaultdict[str, list[str | list[str]]] = defaultdict(list)
    for entry in all_entries():
        authors = authors_from_a01(entry[1])
        for author in authors:
            field_values["author"].append(author)
            if plain_author_re.fullmatch(author) or author in author_corrections().values():
                field_values["plain_author"].append(author)
            else:
                field_values["weird_author"].append(author)

        language = language_from_a01(entry[1])
        if language:
            field_values["language"].append(language)

        title = title_from_a02(entry[2])
        if title:
            field_values["title"].append(title)

        subtitle = subtitle_from_a03(entry[3])
        if subtitle:
            field_values["subtitle"].append(subtitle)

        dewey = dewey_from_a04_a05(entry[4], entry[5])
        if dewey:
            field_values["dewey"].append(dewey)

        entry_numbers = entry_numbers_from_a04_a05_a06_a07_a08_a18_a19(
            entry[4], entry[5], entry[6], entry[7], entry[8], entry[18], entry[19]
        )

        field_values["entry_number_lists"].append(entry_numbers)
        for entry_number in entry_numbers:
            field_values["entry_numbers"].append(entry_number)

        translator = translator_from_a06(entry[6])
        if translator:
            translators = translator.split("!!")
            for single_translator in translators:
                field_values["translator"].append(single_translator)
                translator_surname_name = single_translator.split(",", maxsplit=1)
                if len(translator_surname_name) == 2:
                    field_values["translator_family_name"].append(translator_surname_name[0])
                    if translator_surname_name[1].endswith("."):
                        field_values["translator_name_abbreviations"].append(translator_surname_name[1])
                    else:
                        field_values["translator_names"].append(translator_surname_name[1])

        edition = edition_from_a07(entry[7])
        if edition:
            field_values["edition"].append(edition)

        editor = editor_from_a08_a09(entry[8], entry[9])
        if editor:
            if not editor[0] or not editor[1]:
                field_values["editor"].append(f"{editor[0]} // {editor[1]} ({entry[0]})")
            else:
                field_values["editor"].append(f"{editor[0]} // {editor[1]}")

        edition_year = edition_year_from_a09_a10(entry[9], entry[10])
        if edition_year:
            field_values["edition_year"].append(str(edition_year))

        pages = pages_from_a11(entry[11])
        if pages:
            field_values["pages"].append(str(pages))

        topic_list = topics_from_a12_to_a15_a20_a22_to_a24(
            [
                entry[12],
                entry[13],
                entry[14],
                entry[15],
                entry[20],
                entry[22],
                entry[23],
                entry[24],
            ]
        )
        field_values["topic_lists"].append(topic_list)
        for topic in topic_list:
            field_values["topics"].append(topic)

        curator = curator_from_a16(entry[16])
        if curator:
            field_values["curator"].append(curator)

        copies = copies_from_a17_a18_a30(entry[17], entry[18], entry[30])
        if copies:
            field_values["copies"].append(str(copies))

        donation = donation_from_a17_a30(entry[17], entry[30])
        if donation:
            field_values["donation"].append(donation)

        volume = volume_from_a17_a18_a20_a30(entry[17], entry[18], entry[20], entry[30])
        if volume:
            field_values["volume"].append(volume)

        material = material_from_a18_a30(entry[18], entry[30])
        if material:
            field_values["material"].append(material)

        notes = notes_from_a17_a18_a21_a30(entry[17], entry[18], entry[21], entry[30])
        if notes:
            field_values["notes"].append(notes)

        isbn: Optional[str] = isbn_from_a17_a18_a19_a22_a30(entry[17], entry[18], entry[19], entry[22], entry[30])
        if isbn:
            if not check_isbn(isbn):
                field_values["isbn"].append(isbn)
            if not check_issn(isbn):
                field_values["issn"].append(isbn)
            if not check_ean(isbn):
                field_values["ean"].append(isbn)

    os.makedirs(os.path.join(reports_directory, "calculated-field"), exist_ok=True)

    index = Document()
    index.add_heading("Υπολογισμένες Τιμές")
    links: list[str] = []

    for k, values in field_values.items():
        values = [pprint.pformat(value) for value in values]
        with open(
            os.path.join(reports_directory, "calculated-field", f"calculated_field_{k}.md"), "w", encoding="utf-8"
        ) as outfile:
            doc = Document()
            doc.add_heading(f"Υπολογισμένες τιμές για την ιδιότητα {k}, αλφαβητικά")
            doc.add_unordered_list([f"`{v}`" for v in sorted(set(values))])
            outfile.write(str(doc))
            links.append(
                str(
                    Inline(f"Υπολογισμένες τιμές για την ιδιότητα {k}, αλφαβητικά", link=f"./calculated_field_{k}.html")
                )
            )
        with open(
            os.path.join(reports_directory, "calculated-field", f"calculated_field_{k}_romanize_sort.md"),
            "w",
            encoding="utf-8",
        ) as outfile:
            doc = Document()
            doc.add_heading(f"Υπολογισμένες τιμές για την ιδιότητα {k}, φωνητικά")
            doc.add_unordered_list([f"`{v}` # *{romanize(v)}*" for v in sorted(set(values), key=romanize)])
            outfile.write(str(doc))
            links.append(
                str(
                    Inline(
                        f"Υπολογισμένες τιμές για την ιδιότητα {k}, φωνητικά",
                        link=f"./calculated_field_{k}_romanize_sort.html",
                    )
                )
            )
    index.add_unordered_list(links)
    with open(os.path.join(reports_directory, "calculated-field", "index.md"), "w", encoding="utf-8") as outfile:
        outfile.write(str(index))


def report_invalid_dewey(reports_directory: str):
    invalid_output_dewey: defaultdict[str, list[str]] = defaultdict(list)
    no_output_dewey: defaultdict[str, list[str]] = defaultdict(list)
    for entry in all_entries():
        dewey = dewey_from_a04_a05(entry[4], entry[5])
        if dewey:
            if not is_valid_dewey_strict(dewey):
                invalid_output_dewey[dewey].append(str(entry[0]))
        else:
            if none_if_empty_or_stripped(entry[4]):
                no_output_dewey[entry[4]].append(str(entry[0]))

    os.makedirs(os.path.join(reports_directory, "checks"), exist_ok=True)
    with open(os.path.join(reports_directory, "checks", "invalid_dewey.md"), "w", encoding="utf-8") as outfile:
        doc = Document()
        doc.add_heading("Dewey με προβληματικές τιμές στην έξοδο")
        for k, v in sorted(invalid_output_dewey.items()):
            doc.add_heading(k, level=2)
            doc.add_unordered_list([str(Inline(entry, link=f"../entries/entry_{entry:05}.html")) for entry in v])
        doc.add_heading("Dewey στην είσοδο που δεν βγαίνουν στην έξοδο")
        for k, v in sorted(no_output_dewey.items()):
            doc.add_heading(k, level=2)
            doc.add_unordered_list([str(Inline(entry, link=f"../entries/entry_{entry:05}.html")) for entry in v])
        outfile.write(str(doc))


def report_weird_names(reports_directory: str):
    os.makedirs(os.path.join(reports_directory, "checks"), exist_ok=True)
    weird_translators: list[str] = []
    weird_authors: list[str] = []
    weird_curators: list[str] = []
    weird_donors: list[str] = []
    valid_name_re = re.compile(r"[A-ZΑ-Ω\-]+,[A-ZΑ-Ω\.]*\.?")
    for entry in all_entries():
        translator = translator_from_a06(entry[6])
        if translator:
            translators = translator.split("!!")
            for translator in translators:
                if not valid_name_re.fullmatch(translator):
                    weird_translators.append(translator)
        authors = authors_from_a01(entry[1])
        for author in authors:
            if not valid_name_re.fullmatch(author):
                weird_authors.append(author)
        curator = curator_from_a16(entry[16])
        if curator:
            curators = curator.split("!!")
            for curator in curators:
                if not valid_name_re.fullmatch(curator):
                    weird_curators.append(curator)
        donor = donation_from_a17_a30(entry[17], entry[30])
        if donor:
            donors = donor.split("!!")
            for donor in donors:
                if not valid_name_re.fullmatch(donor):
                    weird_donors.append(donor)

    for field, greek_name, thelist in [
        ("translators", "Μεταφραστές", weird_translators),
        ("authors", "Συγγραφείς", weird_authors),
        ("curators", "Επιμελητές", weird_curators),
        ("donors", "Δωρητές", weird_donors),
    ]:
        with open(os.path.join(reports_directory, "checks", f"invalid_{field}.md"), "w", encoding="utf-8") as outfile:
            doc = Document()
            doc.add_heading(f"{greek_name} με παράξενα ονόματα")
            doc.add_unordered_list(sorted(set(thelist)))
            outfile.write(str(doc))


def report_donors(reports_directory: str):
    os.makedirs(os.path.join(reports_directory, "checks"), exist_ok=True)
    count_map: defaultdict[str, int] = defaultdict(int)
    for entry in all_entries():
        donors = donation_from_a17_a30(entry[17], entry[30])
        if donors:
            for donor in donors.split("!!"):
                count_map[donor] = count_map[donor] + 1
    donor_count_list: list[list[str]] = []
    for donor, count in sorted(count_map.items(), reverse=True, key=lambda x: x[1]):
        donor_count_list.append([donor, str(count)])

    with open(os.path.join(reports_directory, "checks", "donors.md"), "w", encoding="utf-8") as outfile:
        doc = Document()
        doc.add_heading("Δωρητές")
        doc.add_table(["Δωρητής", "Αριθμός βιβλίων"], donor_count_list)
        outfile.write(str(doc))


def report_isbns(reports_directory: str):
    doc = Document()
    doc.add_heading("Προβληματικά ISBN")
    doc.add_table_of_contents()

    for entry in all_entries():
        result = isbn_from_a17_a18_a19_a22_a30(
            entry[17],
            entry[18],
            entry[19],
            entry[22],
            entry[30],
        )
        if not result:
            continue
        checks = [check_isbn(result), check_issn(result), check_ean(result)]
        if None in checks:
            continue

        doc.add_heading(result, level=2)
        doc.add_unordered_list([check for check in checks if check])

        doc.add_heading("Αρχική Καρτέλα στο DBASE", level=3)
        doc.add_code(code=entry_as_yaml(entry, minimal=True), lang="yaml")

    with open(os.path.join(reports_directory, "checks", "invalid_isbn.md"), "w", encoding="utf-8") as outfile:
        outfile.write(str(doc))


def report_entry_numbers(reports_directory: str):
    all_entry_numbers: dict[str, dict[int, str]] = dict()

    no_entry_numbers = Document()
    no_entry_numbers.add_heading("Καρτέλες χωρίς αριθμό εισαγωγής")

    non_numeric = Document()
    non_numeric.add_heading("Καρτέλες με μή αριθμητικό αριθμό εισαγωγής")

    duplicate_entry_numbers: defaultdict[str, list[dict[int, str]]] = defaultdict(list)

    for entry in all_entries():
        entry_numbers = entry_numbers_from_a04_a05_a06_a07_a08_a18_a19(
            entry[4], entry[5], entry[6], entry[7], entry[8], entry[18], entry[19]
        )
        if not entry_numbers:
            no_entry_numbers.add_horizontal_rule()
            no_entry_numbers.add_code(entry_as_yaml(entry, minimal=True), lang="yaml")
        else:
            for n in entry_numbers:
                if n in all_entry_numbers:
                    if not duplicate_entry_numbers[n]:
                        duplicate_entry_numbers[n].append(all_entry_numbers[n])
                    duplicate_entry_numbers[n].append(entry)
                else:
                    all_entry_numbers[n] = entry
                if not n.isnumeric():
                    non_numeric.add_horizontal_rule()
                    non_numeric.add_paragraph(n)
                    non_numeric.add_code(entry_as_yaml(entry, minimal=True), lang="yaml")

    os.makedirs(os.path.join(reports_directory, "checks"), exist_ok=True)
    with open(os.path.join(reports_directory, "checks", "no_entry_numbers.md"), "w", encoding="utf-8") as outfile:
        outfile.write(str(no_entry_numbers))
    with open(
        os.path.join(reports_directory, "checks", "non_numeric_entry_numbers.md"), "w", encoding="utf-8"
    ) as outfile:
        outfile.write(str(non_numeric))

    with open(
        os.path.join(reports_directory, "checks", "duplicate_entry_numbers.md"), "w", encoding="utf-8"
    ) as outfile:
        dup = Document()
        dup.add_heading("Καρτέλες με διπλοπερασμένο αριθμητικό αριθμό εισαγωγής")
        for entry_number, entries in duplicate_entry_numbers.items():
            dup.add_horizontal_rule()
            dup.add_paragraph(entry_number)
            for entry in entries:
                dup.add_code(entry_as_yaml(entry, minimal=True), lang="yaml")
        outfile.write(str(dup))


def entry_as_yaml(entry: dict[int, str], minimal: bool) -> str:
    """return an entry for a code section"""
    if minimal:
        minimal_entry = {k: v for k, v in entry.items() if v}
        return yaml.dump(minimal_entry, default_flow_style=False, allow_unicode=True)
    else:
        return yaml.dump(entry, default_flow_style=False, allow_unicode=True)


def report_entries(reports_directory: str):
    os.makedirs(os.path.join(reports_directory, "entries"), exist_ok=True)
    all: list[tuple[str, str]] = []
    by_author: defaultdict[str, defaultdict[str, list[tuple[str, str]]]] = defaultdict(lambda: defaultdict(list))
    by_dewey: defaultdict[str, list[tuple[str, str]]] = defaultdict(list)

    for entry in all_entries():
        translator = translator_from_a06(entry[6])
        translators: list[str] = []
        if translator:
            for single_translator in translator.split("!!"):
                translator_surname_name = single_translator.split(",", maxsplit=1)
                if len(translator_surname_name) == 2:
                    if translator_surname_name[1].endswith("."):
                        translators.append(
                            f"Επίθετο:{translator_surname_name[0]}, Μή πλήρες όνομα: {translator_surname_name[1]}"
                        )
                    else:
                        translators.append(f"Επίθετο:{translator_surname_name[0]}, Oνομα: {translator_surname_name[1]}")
                else:
                    translators.append(f"{single_translator}")

        donation = donation_from_a17_a30(entry[17], entry[30])
        donors: list[str] = []
        if donation:
            donors = donation.split("!!")

        editor = editor_from_a08_a09(entry[8], entry[9])
        if editor:
            editor = f"{editor[0]} ({editor[1]})"

        topic_list = topics_from_a12_to_a15_a20_a22_to_a24(
            [
                entry[12],
                entry[13],
                entry[14],
                entry[15],
                entry[20],
                entry[22],
                entry[23],
                entry[24],
            ]
        )

        isbn_issn_ean = isbn_from_a17_a18_a19_a22_a30(entry[17], entry[18], entry[19], entry[22], entry[30])
        isbn = ""
        issn = ""
        ean = ""
        if isbn_issn_ean:
            if not check_isbn(isbn_issn_ean):
                isbn = isbn_issn_ean
            if not check_issn(isbn_issn_ean):
                issn = isbn_issn_ean
            if not check_ean(isbn_issn_ean):
                ean = isbn_issn_ean

        doc = Document()

        doc.add_heading("Τίτλος")
        doc.add_paragraph(str(title_from_a02(entry[2])))
        doc.add_paragraph(str(subtitle_from_a03(entry[3])))
        doc.add_paragraph(
            str(
                Inline(
                    text="Στο library.skoufas.gr", link=f"https://library.skoufas.gr/books/by-entry-number/{entry[0]}"
                )
            )
        )

        doc.add_heading("Συγγραφείς", level=2)
        doc.add_unordered_list(authors_from_a01(entry[1]))

        doc.add_heading("Αριθμοί Εισαγωγης", level=2)
        doc.add_unordered_list(
            entry_numbers_from_a04_a05_a06_a07_a08_a18_a19(
                entry[4], entry[5], entry[6], entry[7], entry[8], entry[18], entry[19]
            )
        )

        doc.add_table(
            ["Πεδίο", "Τιμή"],
            [
                ["dbase_number", str(entry[0])],
                ["Γλώσσα", str(language_from_a01(entry[1]))],
                ["Dewey", str(dewey_from_a04_a05(entry[4], entry[5]))],
                ["Έκδοση", str(edition_from_a07(entry[7]))],
                ["Εκδότης (Πόλη)", f"{editor}"],
                ["Χρόνος έκδοσης", f"{edition_year_from_a09_a10(entry[9], entry[10])}"],
                ["Σελίδες", f"{pages_from_a11(entry[11])}"],
                ["Επιμελητής", f"{curator_from_a16(entry[16])}"],
                ["Αντίτυπα", f"{copies_from_a17_a18_a30(entry[17], entry[18], entry[30])}"],
                ["Δωρητές", str(MDList(donors))],
                ["Τεύχος/Τόμος", f"{volume_from_a17_a18_a20_a30(entry[17], entry[18], entry[20], entry[30])}"],
                ["Υλικό", f"{material_from_a18_a30(entry[18], entry[30])}"],
                ["Σημειώσεις", f"{notes_from_a17_a18_a21_a30(entry[17], entry[18], entry[21], entry[30])}"],
                ["ISBN", isbn],
                ["ISSN", issn],
                ["EAN", ean],
            ],
            [Table.Align.LEFT, Table.Align.RIGHT],
            0,
        )

        doc.add_paragraph("Μεταφραστές")
        doc.add_unordered_list(translators)

        doc.add_paragraph("Θέματα")
        doc.add_unordered_list(topic_list)

        doc.add_paragraph("Ιδιότητες")
        doc.add_block(
            MDList(
                ["Εχει CD"],
                checked=has_cd_from_a02_a03_a12_a13_a14_a17_a18_a22_a30(
                    [
                        entry[2],
                        entry[3],
                        entry[12],
                        entry[13],
                        entry[14],
                        entry[17],
                        entry[18],
                        entry[22],
                        entry[30],
                    ]
                ),
            )
        )

        doc.add_block(
            MDList(
                ["Εχει DVD"],
                checked=has_dvd_from_a30(
                    [
                        entry[30],
                    ]
                ),
            )
        )

        doc.add_block(MDList(["Ανάτυπο"], checked=offprint_from_a17_a21_a30(entry[17], entry[21], entry[30])))

        doc.add_heading("Αρχική Καρτέλα στο DBASE", level=1)
        doc.add_code(code=entry_as_yaml(entry, minimal=False), lang="yaml")

        title = title_from_a02(entry[2])
        if not title:
            title = "Χωρίς Τίτλο"
        subtitle = subtitle_from_a03(entry[3])
        if subtitle:
            title += " - " + subtitle

        all.append((entry[0], title))
        for author in authors_from_a01(entry[1]):
            if len(author) == 0:
                by_author["#"]["Χωρίς συγγραφέα"].append((entry[0], title))
            else:
                by_author[author[0]][author].append((entry[0], title))

        dewey = dewey_from_a04_a05(entry[4], entry[5])

        if dewey:
            by_dewey[dewey].append((entry[0], title))
        else:
            by_dewey["Xωρίς dewey"].append((entry[0], title))

        with open(
            os.path.join(reports_directory, "entries", f"entry_{entry[0]:05}.md"), "w", encoding="utf-8"
        ) as outfile:
            outfile.write(str(doc))

    index = Document()
    index.add_heading("Όλες οι καρτέλες")
    index.add_table_of_contents(range(1, 3))

    def divide_chunks(inlist: list[tuple[str, str]], size: int):
        for i in range(0, len(inlist), size):
            yield inlist[i : i + size]

    all_divided = list(divide_chunks(all, 1000))
    for thousands, sublist in enumerate(all_divided):
        index.add_heading(f"Απο {thousands*1000} ως {thousands*1000+len(sublist)}")
        sublist = [str(Inline(f"{int(id):05}: {title}", link=f"./entry_{int(id):05}.html")) for id, title in sublist]
        index.add_unordered_list(sublist)

    with open(os.path.join(reports_directory, "entries", "index.md"), "w", encoding="utf-8") as outfile:
        outfile.write(str(index))

    index_by_author = Document()
    index_by_author.add_heading("Όλες οι καρτέλες, κατα συγγραφέα")
    index_by_author.add_table_of_contents(range(1, 3))
    for author_initial, author_dict in sorted(by_author.items()):
        index_by_author.add_heading(author_initial, level=2)
        for author, entry_list in sorted(author_dict.items()):
            index_by_author.add_heading(author, level=3)
            entry_list = [
                str(Inline(f"{int(id):05}: {title}", link=f"./entry_{int(id):05}.html")) for id, title in entry_list
            ]
            index_by_author.add_unordered_list(entry_list)
    with open(os.path.join(reports_directory, "entries", "index_by_author.md"), "w", encoding="utf-8") as outfile:
        outfile.write(str(index_by_author))

    index_by_dewey = Document()
    index_by_dewey.add_heading("Όλες οι καρτέλες, κατα συγγραφέα")
    index_by_dewey.add_table_of_contents(range(1, 3))
    for dewey, entry_list in sorted(by_dewey.items()):
        index_by_dewey.add_heading(dewey, level=2)
        entry_list = [
            str(Inline(f"{int(id):05}: {title}", link=f"./entry_{int(id):05}.html")) for id, title in entry_list
        ]
        index_by_dewey.add_unordered_list(entry_list)
    with open(os.path.join(reports_directory, "entries", "index_by_dewey.md"), "w", encoding="utf-8") as outfile:
        outfile.write(str(index_by_dewey))


def add_index(reports_directory: str):
    os.makedirs(reports_directory, exist_ok=True)
    doc = Document()

    doc.add_heading("Βιβλιοθήκη Σκουφά: προσωρινός κατάλογος βιβλίων")
    doc.add_table_of_contents()

    doc.add_paragraph(str(Inline("Όλες οι καρτέλες", link="./entries/index.html")))
    doc.add_paragraph(str(Inline("Όλες οι καρτέλες, κατα συγγραφέα", link="./entries/index_by_author.html")))
    doc.add_paragraph(str(Inline("Όλες οι καρτέλες, κατα dewey", link="./entries/index_by_dewey.html")))

    doc.add_heading("Τιμές κατα στήλη")
    doc.add_paragraph(str(Inline("Υπολογισμένες τιμές", link="./calculated-field/index.html")))
    doc.add_paragraph(str(Inline("Τιμές στις στήλες των καρτελών", link="./single-field/index.html")))

    doc.add_heading("Ελεγκτικές Αναφορές")
    doc.add_paragraph(
        str(
            Inline(
                "Καρτέλες με διπλοπερασμένο αριθμητικό αριθμό εισαγωγής", link="./checks/duplicate_entry_numbers.html"
            )
        )
    )
    doc.add_paragraph(str(Inline("Καρτέλες χωρίς αριθμό εισαγωγής", link="./checks/no_entry_numbers.html")))
    doc.add_paragraph(
        str(Inline("Καρτέλες με μή αριθμητικό αριθμό εισαγωγής", link="./checks/non_numeric_entry_numbers.html"))
    )
    doc.add_paragraph(str(Inline("Dewey με προβληματικές τιμές", link="./checks/invalid_dewey.html")))

    doc.add_paragraph(str(Inline("Μεταφραστές με παράξενα ονόματα", link="./checks/invalid_translators.html")))
    doc.add_paragraph(str(Inline("Συγγραφείς με παράξενα ονόματα", link="./checks/invalid_authors.html")))
    doc.add_paragraph(str(Inline("Επιμελητές με παράξενα ονόματα", link="./checks/invalid_curators.html")))
    doc.add_paragraph(str(Inline("Δωρητές με παράξενα ονόματα", link="./checks/invalid_donors.html")))

    doc.add_paragraph(str(Inline("Προβληματικά ISBN", link="./checks/invalid_isbn.html")))

    doc.add_paragraph(str(Inline("Δωρητές", link="./checks/donors.html")))

    with open(os.path.join(reports_directory, "index.md"), "w", encoding="utf-8") as outfile:
        outfile.write(str(doc))


def main():
    """Create markdown reports"""
    if len(sys.argv) > 0:
        md_report_dir = os.path.abspath(sys.argv[1])
    else:
        md_report_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir, os.pardir, "md_reports")
    print(f"Creating reports in {md_report_dir}")
    shutil.rmtree(md_report_dir, ignore_errors=True)
    add_index(md_report_dir)
    report_weird_names(md_report_dir)
    report_donors(md_report_dir)
    report_isbns(md_report_dir)
    report_entry_numbers(md_report_dir)
    report_entries(md_report_dir)
    report_single_fields(md_report_dir)
    report_single_extracted_fields(md_report_dir)
    report_invalid_dewey(md_report_dir)
    print(f"Finished creating reports in {md_report_dir}")


if __name__ == "__main__":
    main()



""" Functions that extract information given specific strings """
import re
from functools import cache
from typing import Optional, Any

from skoufas_dbf_reader.utilities import read_yaml_data


def has_author(a01: Optional[str]) -> bool:
    """Check values for marks of missing author"""
    if not a01:
        return False
    if a01 in author_corrections() and not author_corrections().get(a01):
        return False
    return True


@cache
def converted_entries() -> Any:
    """All converted entries"""
    return read_yaml_data("converted_entries")


@cache
def language_codes() -> dict[str, str]:
    """Map of language codes in A01 to ISO language codes"""
    return read_yaml_data("language_codes")


@cache
def author_corrections() -> dict[str, Optional[str]]:
    """Map of author names found and manual overrides"""
    return read_yaml_data("author_corrections")


@cache
def editor_corrections() -> dict[str, str]:
    """Map editor and place manual overrides"""
    return read_yaml_data("editor_corrections")


@cache
def field04_corrections() -> dict[str, Optional[str | dict[str, str]]]:
    """Map of invalid dewey codes found and manual overrides"""
    return read_yaml_data("field04_corrections")


@cache
def field05_corrections() -> dict[str, Optional[str | dict[str, str | bool]]]:
    """Map of invalid entry numbers found and manual overrides"""
    return read_yaml_data("field05_corrections")


@cache
def field06_corrections() -> dict[str, Optional[str | dict[str, str | bool]]]:
    """Map of invalid entry numbers found and manual overrides"""
    return read_yaml_data("field06_corrections")


@cache
def field07_corrections() -> dict[str, Optional[str | dict[str, str | bool]]]:
    """Map of entry numbers and manual overrides"""
    return read_yaml_data("field07_corrections")


@cache
def field08_corrections() -> dict[str, Optional[str | dict[str, str]]]:
    """Map of editors and manual overrides"""
    return read_yaml_data("field08_corrections")


@cache
def field09_corrections() -> dict[str, Optional[str | dict[str, str]]]:
    """Map of editor places and manual overrides"""
    return read_yaml_data("field09_corrections")


@cache
def field10_corrections() -> dict[str, Optional[str]]:
    """Map of year and manual overrides"""
    return read_yaml_data("field10_corrections")


@cache
def field11_corrections() -> dict[str, Optional[str]]:
    """Map of pages and manual overrides"""
    return read_yaml_data("field11_corrections")


@cache
def field16_corrections() -> dict[str, Optional[str]]:
    """Map of curators and manual overrides"""
    return read_yaml_data("field16_corrections")


@cache
def field17_corrections() -> dict[str, Optional[str | dict[str, str | bool | int]]]:
    """Map of manual overrides"""
    return read_yaml_data("field17_corrections")


@cache
def field18_corrections() -> dict[str, Optional[str | dict[str, str | bool | int]]]:
    """Map of manual overrides"""
    return read_yaml_data("field18_corrections")


@cache
def field19_corrections() -> dict[str, Optional[str | dict[str, str]]]:
    """Map of manual overrides"""
    return read_yaml_data("field19_corrections")


@cache
def field20_corrections() -> dict[str, Optional[str | dict[str, str | bool | int]]]:
    """Map of manual overrides"""
    return read_yaml_data("field20_corrections")


@cache
def field30_corrections() -> dict[str, Optional[str | dict[str, str | bool | int]]]:
    """Map of manual overrides"""
    return read_yaml_data("field30_corrections")


@cache
def topic_replacements() -> dict[str, Optional[str]]:
    """Map of topic name manual overrides"""
    return read_yaml_data("topic_replacements")


@cache
def translator_corrections() -> dict[str, str]:
    """Map of translator names found and manual overrides"""
    return read_yaml_data("translator_corrections")


dewey_re1 = [
    re.compile(r"([0-9]{3})"),
    re.compile(r"([0-9]{3}\.[0-9]+)"),
]
dewey_re2 = [
    re.compile(r"([0-9]{3}\.[0-9]+)\s+([^0-9\.]*)"),
    re.compile(r"([0-9]{3}\.[0-9]+)([^0-9\.]*)"),
    re.compile(r"([0-9]{3})\s+([^0-9\.]*)"),
    re.compile(r"([0-9]{3})([^0-9\.]*)"),
]

valid_pages_re = re.compile(r"(\d+)\s*(Σ|S|ΣΕΛ|Δ|Σ Ρ|ΣΑ|ΣΙΣ|Σ Ε|ΣΕΓ|Σ18|ΣΚΑ|ΣΑΜ|Σ Ο|s|Σ Λ|Σ  Α|Σ11|Σ#Ξ|Σ Ι|σ|Φ)*")

topic_in_paren_re = re.compile(r".*?\((.*)\).*")

has_cd_re = re.compile(r"\bCD\b", re.IGNORECASE)

has_dvd_re = re.compile(r"\bDVD\b", re.IGNORECASE)

a22_has_isbn_part_re = re.compile(r"[0-9\-]+")

plain_author_re = re.compile(r"[A-ZΑ-Ω0-1]+,?[A-ZΑ-Ω0-1]*\.?")



from __future__ import annotations

from skoufas_dbf_reader.correction_data import converted_entries
from skoufas_dbf_reader.utilities import all_entries, none_if_empty_or_stripped, read_yaml_data, romanize


def test_yaml_data():
    no_author = read_yaml_data("language_codes")
    assert no_author
    assert type(no_author) == dict
    assert len(no_author) > 0


def test_none_if_empty_or_stripped():
    assert none_if_empty_or_stripped(None) is None
    assert none_if_empty_or_stripped("") is None
    assert none_if_empty_or_stripped(" ") is None
    assert none_if_empty_or_stripped("   ") is None
    assert none_if_empty_or_stripped("foo") == "foo"
    assert none_if_empty_or_stripped("    foo ") == "foo"


def test_all_entries():
    entries = all_entries()
    assert entries
    assert type(entries) == list
    assert len(entries) > 0
    entry = entries[0]
    assert type(entry) == dict
    assert entry[0] == 1

    assert entries[1000][0] == 1001


def test_converted_entries():
    all = converted_entries()
    assert all


def test_romanize():
    assert romanize("Γειά") == "Geia"
    assert romanize("") == ""
    assert romanize(None) == ""
